from datetime import datetime
from typing import Any, List, TypedDict

from ..enums import HDRSeverity, HDRStatus


class HDREvent(TypedDict):
    time: datetime
    msg: str
    code: str
    severity: HDRSeverity


class HDRResult(TypedDict):
    status: HDRStatus
    value: Any
    events: List[HDREvent]
