"""
Netlab enums.
"""

from .common import DateFormat, HDRSeverity, HDRStatus, TimeFormat
from .klass import ClassEmailLogs, ClassExtensionSlots, ClassLabLimit, ContentAccessiblity
from .pod import (
    PCIcon,
    PCType,
    PLIcon,
    PodAdminState,
    PodCategory,
    PodCurrentState,
    PodState,
    RemoveVMS,
    VirtualHostComPath,
    VirtualMachineRole,
)
from .reservation import InitConfig, ReservationScope, ReservationType
from .system import FreeAlarm
from .user import AccountPrivileges, AccountType, CommunityExtensionSlots
from .vm import CloneStorageAllocation, CloneType, PowerState, VirtualMachineLicenseType

__all__ = (
    "TimeFormat",
    "DateFormat",
    "HDRSeverity",
    "HDRStatus",
    "ClassEmailLogs",
    "ClassLabLimit",
    "ClassExtensionSlots",
    "PCIcon",
    "PCType",
    "PLIcon",
    "PodCategory",
    "PodState",
    "RemoveVMS",
    "VirtualHostComPath",
    "VirtualMachineRole",
    "PodAdminState",
    "PodCurrentState",
    "ReservationType",
    "ReservationScope",
    "InitConfig",
    "FreeAlarm",
    "AccountPrivileges",
    "AccountType",
    "CommunityExtensionSlots",
    "ContentAccessiblity",
    "PowerState",
    "CloneStorageAllocation",
    "CloneType",
    "VirtualMachineLicenseType",
)
